@@ -63,7 +63,7 @@ vivso_find(uint32_t iana_en, const void *arg)
 	return NULL;
 }
 
-size_t
+ssize_t
 dhcp_vendor(char *str, size_t len)
 {
 	struct utsname utn;
@@ -71,17 +71,21 @@ dhcp_vendor(char *str, size_t len)
 	int l;
 
 	if (uname(&utn) != 0)
-		return (size_t)snprintf(str, len, "%s-%s",
+		return (ssize_t)snprintf(str, len, "%s-%s",
 		    PACKAGE, VERSION);
 	p = str;
 	l = snprintf(p, len,
 	    "%s-%s:%s-%s:%s", PACKAGE, VERSION,
 	    utn.sysname, utn.release, utn.machine);
+	if (l == -1 || (size_t)(l + 1) > len)
+		return -1;
 	p += l;
 	len -= (size_t)l;
 	l = if_machinearch(p, len);
+	if (l == -1 || (size_t)(l + 1) > len)
+		return -1;
 	p += l;
-	return (size_t)(p - str);
+	return p - str;
 }
 
 int
@@ -141,11 +145,9 @@ make_option_mask(const struct dhcp_opt *dopts, size_t dopts_len,
 			return -1;
 		}
 		if (add == 1 || add == 2)
-			add_option_mask(mask,
-			    opt->option);
+			add_option_mask(mask, opt->option);
 		else
-			del_option_mask(mask,
-			    opt->option);
+			del_option_mask(mask, opt->option);
 	}
 	free(o);
 	return 0;
@@ -231,6 +233,10 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 			if (ltype == 0x80 || ltype == 0x40)
 				return -1;
 			else if (ltype == 0xc0) { /* pointer */
+				if (q == e) {
+					errno = ERANGE;
+					return -1;
+				}
 				l = (l & 0x3f) << 8;
 				l |= *q++;
 				/* save source of first jump. */
@@ -248,6 +254,10 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 				}
 			} else {
 				/* straightforward name segment, add with '.' */
+				if (q + l > e) {
+					errno = ERANGE;
+					return -1;
+				}
 				count += l + 1;
 				if (out) {
 					if (l + 1 > len) {
