@@ -123,7 +123,7 @@ struct udp_dhcp_packet
 
 static const size_t udp_dhcp_len = sizeof(struct udp_dhcp_packet);
 
-static int dhcp_open(struct interface *);
+static int dhcp_open(struct interface *ifp);
 
 void
 dhcp_printoptions(const struct dhcpcd_ctx *ctx,
@@ -299,6 +299,10 @@ decode_rfc3442(char *out, size_t len, const uint8_t *p, size_t pl)
 			return -1;
 		}
 		ocets = (cidr + 7) / NBBY;
+		if (p + 4 + ocets > e) {
+			errno = ERANGE;
+			return -1;
+		}
 		if (!out) {
 			p += 4 + ocets;
 			bytes += ((4 * 4) * 2) + 4;
@@ -361,6 +365,13 @@ decode_rfc3442_rt(const uint8_t *data, size_t dl)
 			return NULL;
 		}
 
+		ocets = (cidr + 7) / NBBY;
+		if (p + 4 + ocets > e) {
+			ipv4_freeroutes(routes);
+			errno = ERANGE;
+			return NULL;
+		}
+
 		rt = calloc(1, sizeof(*rt));
 		if (rt == NULL) {
 			syslog(LOG_ERR, "%s: %m", __func__);
@@ -369,7 +380,6 @@ decode_rfc3442_rt(const uint8_t *data, size_t dl)
 		}
 		TAILQ_INSERT_TAIL(routes, rt, next);
 
-		ocets = (cidr + 7) / NBBY;
 		/* If we have ocets then we have a destination and netmask */
 		if (ocets > 0) {
 			memcpy(&rt->dest.s_addr, p, ocets);
@@ -828,7 +838,6 @@ make_message(struct dhcp_message **message,
 			p += ifo->vendorclassid[0] + 1;
 		}
 
-
 		if (type != DHCP_INFORM) {
 			if (ifo->leasetime != 0) {
 				*p++ = DHO_LEASETIME;
@@ -990,7 +999,11 @@ make_message(struct dhcp_message **message,
 		auth_len = (size_t)dhcp_auth_encode(&ifo->auth,
 		    state->auth.token,
 		    NULL, 0, 4, type, NULL, 0);
-		if (auth_len > 0) {
+		if ((ssize_t)auth_len == -1) {
+			syslog(LOG_ERR, "%s: dhcp_auth_encode: %m",
+			    iface->name);
+			auth_len = 0;
+		} else if (auth_len != 0) {
 			len = (size_t)((p + auth_len) - m);
 			if (auth_len > 255 || len > sizeof(*dhcp))
 				goto toobig;
@@ -998,9 +1011,7 @@ make_message(struct dhcp_message **message,
 			*p++ = (uint8_t)auth_len;
 			auth = p;
 			p += auth_len;
-		} else if ((ssize_t)auth_len == -1)
-			syslog(LOG_ERR, "%s: dhcp_auth_encode: %m",
-			    iface->name);
+		}
 	}
 
 	*p++ = DHO_END;
@@ -1014,7 +1025,7 @@ make_message(struct dhcp_message **message,
 #endif
 
 	len = (size_t)(p - m);
-	if (ifo->auth.options & DHCPCD_AUTH_SEND && auth_len > 0)
+	if (ifo->auth.options & DHCPCD_AUTH_SEND && auth_len != 0)
 		dhcp_auth_encode(&ifo->auth, state->auth.token,
 		    m, len, 4, type, auth, auth_len);
 
@@ -1395,26 +1406,21 @@ dhcp_close(struct interface *ifp)
 		return;
 
 	if (state->arp_fd != -1) {
-		eloop_event_delete(ifp->ctx->eloop, state->arp_fd);
+		eloop_event_delete(ifp->ctx->eloop, state->arp_fd, 0);
 		close(state->arp_fd);
 		state->arp_fd = -1;
 	}
 	if (state->raw_fd != -1) {
-		eloop_event_delete(ifp->ctx->eloop, state->raw_fd);
+		eloop_event_delete(ifp->ctx->eloop, state->raw_fd, 0);
 		close(state->raw_fd);
 		state->raw_fd = -1;
 	}
-	if (state->udp_fd != -1) {
-		eloop_event_delete(ifp->ctx->eloop, state->udp_fd);
-		close(state->udp_fd);
-		state->udp_fd = -1;
-	}
 
 	state->interval = 0;
 }
 
 static int
-dhcp_openudp(struct dhcpcd_ctx *ctx, struct interface *ifp)
+dhcp_openudp(struct interface *ifp)
 {
 	int s;
 	struct sockaddr_in sin;
@@ -1466,31 +1472,13 @@ dhcp_openudp(struct dhcpcd_ctx *ctx, struct interface *ifp)
 	if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) == -1)
 		goto eexit;
 
-	if (ifp)
-		state->udp_fd = s;
-	else
-		ctx->udp_fd = s;
-	return 0;
+	return s;
 
 eexit:
 	close(s);
 	return -1;
 }
 
-static ssize_t
-dhcp_sendpacket(const struct interface *iface, struct in_addr to,
-    const uint8_t *data, size_t len)
-{
-	struct sockaddr_in sin;
-
-	memset(&sin, 0, sizeof(sin));
-	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = to.s_addr;
-	sin.sin_port = htons(DHCP_SERVER_PORT);
-	return sendto(D_CSTATE(iface)->udp_fd, data, len, 0,
-	    (struct sockaddr *)&sin, sizeof(sin));
-}
-
 static uint16_t
 checksum(const void *data, uint16_t len)
 {
@@ -1575,8 +1563,9 @@ send_message(struct interface *iface, uint8_t type,
 	size_t len;
 	ssize_t r;
 	struct in_addr from, to;
-	in_addr_t a = 0;
+	in_addr_t a = INADDR_ANY;
 	struct timeval tv;
+	int s;
 
 	if (!callback)
 		syslog(LOG_DEBUG, "%s: sending %s with xid 0x%x",
@@ -1599,23 +1588,30 @@ send_message(struct interface *iface, uint8_t type,
 		    timeval_to_double(&tv));
 	}
 
-	/* Ensure sockets are open. */
-	if (dhcp_open(iface) == -1) {
-		if (!(iface->ctx->options & DHCPCD_TEST))
-			dhcp_drop(iface, "FAIL");
+	if (dhcp_open(iface) == -1)
 		return;
-	}
+
+	if (state->addr.s_addr != INADDR_ANY &&
+	    state->new != NULL &&
+	    (state->new->cookie == htonl(MAGIC_COOKIE) ||
+	    iface->options->options & DHCPCD_INFORM))
+	{
+		s = dhcp_openudp(iface);
+		if (s == -1 && errno != EADDRINUSE)
+			syslog(LOG_ERR, "%s: dhcp_openudp: %m", iface->name);
+	} else
+		s = -1;
 
 	/* If we couldn't open a UDP port for our IP address
 	 * then we cannot renew.
 	 * This could happen if our IP was pulled out from underneath us.
 	 * Also, we should not unicast from a BOOTP lease. */
-	if (state->udp_fd == -1 ||
+	if (s == -1 ||
 	    (!(ifo->options & DHCPCD_INFORM) &&
 	    is_bootp(iface, state->new)))
 	{
 		a = state->addr.s_addr;
-		state->addr.s_addr = 0;
+		state->addr.s_addr = INADDR_ANY;
 	}
 	r = make_message(&dhcp, iface, type);
 	if (r == -1)
@@ -1627,13 +1623,18 @@ send_message(struct interface *iface, uint8_t type,
 	if (from.s_addr)
 		to.s_addr = state->lease.server.s_addr;
 	else
-		to.s_addr = 0;
+		to.s_addr = INADDR_ANY;
 	if (to.s_addr && to.s_addr != INADDR_BROADCAST) {
-		r = dhcp_sendpacket(iface, to, (uint8_t *)dhcp, len);
-		if (r == -1) {
+		struct sockaddr_in sin;
+
+		memset(&sin, 0, sizeof(sin));
+		sin.sin_family = AF_INET;
+		sin.sin_addr.s_addr = to.s_addr;
+		sin.sin_port = htons(DHCP_SERVER_PORT);
+		r = sendto(s, (uint8_t *)dhcp, len, 0,
+		    (struct sockaddr *)&sin, sizeof(sin));
+		if (r == -1)
 			syslog(LOG_ERR, "%s: dhcp_sendpacket: %m", iface->name);
-			dhcp_close(iface);
-		}
 	} else {
 		size_t ulen;
 
@@ -1672,6 +1673,9 @@ send_message(struct interface *iface, uint8_t type,
 	free(dhcp);
 
 fail:
+	if (s != -1)
+		close(s);
+
 	/* Even if we fail to send a packet we should continue as we are
 	 * as our failure timeouts will change out codepath when needed. */
 	if (callback)
@@ -2141,7 +2145,7 @@ dhcp_drop(struct interface *ifp, const char *reason)
 	dhcp_auth_reset(&state->auth);
 	dhcp_close(ifp);
 	arp_close(ifp);
-	eloop_timeouts_delete(ifp->ctx->eloop, ifp, dhcp_expire, NULL);
+	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 	if (ifp->options->options & DHCPCD_RELEASE) {
 		unlink(state->leasefile);
 		if (ifp->carrier != LINK_DOWN &&
@@ -2693,39 +2697,22 @@ dhcp_handlepacket(void *arg)
 	free(dhcp);
 }
 
-static void
-dhcp_handleudp1(struct dhcpcd_ctx *ctx, int *fd, const char *ifname)
-{
-	uint8_t buffer[sizeof(struct dhcp_message)];
-
-	/* Just read what's in the UDP fd and discard it as we always read
-	 * from the raw fd */
-	if (read(*fd, buffer, sizeof(buffer)) == -1) {
-		syslog(LOG_ERR, "%s: %s: %m", ifname, __func__);
-		eloop_event_delete(ctx->eloop, *fd);
-		close(*fd);
-		*fd = -1;
-	}
-}
-
 static void
 dhcp_handleudp(void *arg)
 {
 	struct dhcpcd_ctx *ctx;
+	uint8_t buffer[sizeof(struct dhcp_message)];
 
 	ctx = arg;
-	dhcp_handleudp1(arg, &ctx->udp_fd, NULL);
-}
 
-static void
-dhcp_handleifudp(void *arg)
-{
-	const struct interface *ifp;
-	struct dhcp_state *state;
-
-	ifp = arg;
-	state = D_STATE(ifp);
-	dhcp_handleudp1(ifp->ctx, &state->udp_fd, ifp->name);
+	/* Just read what's in the UDP fd and discard it as we always read
+	 * from the raw fd */
+	if (read(ctx->udp_fd, buffer, sizeof(buffer)) == -1) {
+		syslog(LOG_ERR, "%s: %m", __func__);
+		eloop_event_delete(ctx->eloop, ctx->udp_fd, 0);
+		close(ctx->udp_fd);
+		ctx->udp_fd = -1;
+	}
 }
 
 static int
@@ -2749,21 +2736,7 @@ dhcp_open(struct interface *ifp)
 			return -1;
 		}
 		eloop_event_add(ifp->ctx->eloop,
-		    state->raw_fd, dhcp_handlepacket, ifp);
-	}
-	if (state->udp_fd == -1 &&
-	    state->addr.s_addr != 0 &&
-	    state->new != NULL &&
-	    (state->new->cookie == htonl(MAGIC_COOKIE) ||
-	    ifp->options->options & DHCPCD_INFORM))
-	{
-		if (dhcp_openudp(ifp->ctx, ifp) == -1 && errno != EADDRINUSE) {
-			syslog(LOG_ERR, "%s: dhcp_openudp: %m", ifp->name);
-			return -1;
-		}
-		if (state->udp_fd != -1)
-			eloop_event_add(ifp->ctx->eloop,
-			    state->udp_fd, dhcp_handleifudp, ifp);
+		    state->raw_fd, dhcp_handlepacket, ifp, NULL, NULL);
 	}
 	return 0;
 }
@@ -2823,7 +2796,7 @@ dhcp_free(struct interface *ifp)
 	}
 	if (ifp == NULL) {
 		if (ctx->udp_fd != -1) {
-			eloop_event_delete(ctx->eloop, ctx->udp_fd);
+			eloop_event_delete(ctx->eloop, ctx->udp_fd, 0);
 			close(ctx->udp_fd);
 			ctx->udp_fd = -1;
 		}
@@ -2850,7 +2823,7 @@ dhcp_init(struct interface *ifp)
 		if (state == NULL)
 			return -1;
 		/* 0 is a valid fd, so init to -1 */
-		state->raw_fd = state->udp_fd = state->arp_fd = -1;
+		state->raw_fd = state->arp_fd = -1;
 	}
 
 	state->state = DHS_INIT;
@@ -2931,9 +2904,15 @@ dhcp_start1(void *arg)
 
 	/* Listen on *.*.*.*:bootpc so that the kernel never sends an
 	 * ICMP port unreachable message back to the DHCP server */
-	if (ifp->ctx->udp_fd == -1 && dhcp_openudp(ifp->ctx, NULL) != -1)
+	if (ifp->ctx->udp_fd == -1) {
+		ifp->ctx->udp_fd = dhcp_openudp(NULL);
+		if (ifp->ctx->udp_fd == -1) {
+			syslog(LOG_ERR, "dhcp_openudp: %m");
+			return;
+		}
 		eloop_event_add(ifp->ctx->eloop,
-		    ifp->ctx->udp_fd, dhcp_handleudp, ifp->ctx);
+		    ifp->ctx->udp_fd, dhcp_handleudp, ifp->ctx, NULL, NULL);
+	}
 
 	if (dhcp_init(ifp) == -1) {
 		syslog(LOG_ERR, "%s: dhcp_init: %m", ifp->name);
@@ -3035,20 +3014,13 @@ dhcp_start(struct interface *ifp)
 		return;
 
 	/* No point in delaying a static configuration */
-	if (ifp->options->options & DHCPCD_STATIC &&
-	    !(ifp->options->options & DHCPCD_INFORM))
-	{
-		tv.tv_sec = 0;
-		tv.tv_usec = 0;
-	} else {
-		tv.tv_sec = DHCP_MIN_DELAY;
-		tv.tv_usec = (suseconds_t)arc4random_uniform(
-		    (DHCP_MAX_DELAY - DHCP_MIN_DELAY) * 1000000);
-		timernorm(&tv);
-		syslog(LOG_DEBUG,
-		    "%s: delaying DHCP for %0.1f seconds",
-		    ifp->name, timeval_to_double(&tv));
-	}
+	tv.tv_sec = DHCP_MIN_DELAY;
+	tv.tv_usec = (suseconds_t)arc4random_uniform(
+	    (DHCP_MAX_DELAY - DHCP_MIN_DELAY) * 1000000);
+	timernorm(&tv);
+	syslog(LOG_DEBUG,
+	    "%s: delaying IPv4 for %0.1f seconds",
+	    ifp->name, timeval_to_double(&tv));
 
 	eloop_timeout_add_tv(ifp->ctx->eloop, &tv, dhcp_start1, ifp);
 }
