@@ -64,7 +64,8 @@ struct interface {
 	unsigned int metric;
 	int carrier;
 	int wireless;
-	char ssid[IF_SSIDSIZE];
+	uint8_t ssid[IF_SSIDSIZE];
+	unsigned int ssid_len;
 
 	char profile[PROFILE_LEN];
 	struct if_options *options;
@@ -95,7 +96,8 @@ struct dhcpcd_ctx {
 	struct eloop_ctx *eloop;
 
 	int control_fd;
-	struct fd_list *control_fds;
+	int control_unpriv_fd;
+	struct fd_list_head control_fds;
 	char control_sock[sizeof(CONTROLSOCKET) + IF_NAMESIZE];
 	gid_t control_group;
 
@@ -140,6 +142,8 @@ struct dhcpcd_ctx {
 extern const int dhcpcd_handlesigs[];
 #endif
 
+int dhcpcd_oneup(struct dhcpcd_ctx *);
+int dhcpcd_ipwaited(struct dhcpcd_ctx *);
 pid_t dhcpcd_daemonise(struct dhcpcd_ctx *);
 
 int dhcpcd_handleargs(struct dhcpcd_ctx *, struct fd_list *, int, char **);
